/*
 * fd_forms.c
 *
 * This file is part of the Forms Designer.
 *
 * It contains the routines that maintain the collection of
 * forms on which the program is working. It contains the callback
 * routines to add forms, change their name, remove them ,etc.
 * It also contains the routine to draw them and the  basic routines
 * for loading and saving forms.
 *
 * Written by Mark Overmars
 *
 * Version 2.1a
 * Date  Jan 20, 1992
 */

#include <string.h>
#include <stdio.h>
#include <gl/gl.h>
#include "fd_main.h"
#include "fd_theforms.h"

/**** 
   THE DATA STRUCTURE 
****/

typedef struct {
  FL_FORM *form;	/* The form */
  char fname[40];	/* Its name */
} FRM;

static FRM forms[MAXFRM];	/* The forms */
int fnumb = 0;

FL_FORM *cur_form = NULL;	/* The current form */

static int get_form_numb(FL_FORM *form)
/* Returns the number of the form */
{
  int i;
  for (i=0; i<fnumb; i++) if (forms[i].form == form) return i;
  return -1;
}

static void set_form(int numb)
/* Sets the current form to numb, to NULL when numb = -1. */
{
  if (numb == -1)
  {
    cur_form = NULL;
    fl_deselect_browser(formbrowser);
  }
  else
  {
    cur_form = forms[numb].form;
    set_bounding_box(0.0,0.0,cur_form->w,cur_form->h);
    fl_select_browser_line(formbrowser,numb+1);
  }
  clear_selection();
  fillin_groups();
  redraw_the_form(1);
}

/****
  CALLBACK ROUTINES
****/

void form_cb(FL_OBJECT *obj, long arg)
/* Callback routine that is called when the user selects another form
   to work on. */
  { set_form(fl_get_browser(formbrowser)-1); }

void addform_cb(FL_OBJECT *obj, long arg)
/* Callback routine called when the user adds a form. */
{
  float x = 0.0, y = 0.0, w, h;
  long xx,yy;
  if (fnumb == MAXFRM) return;
  /* Make old form invisible */
  set_form(-1);
  /* Get boundary */
  fl_deactivate_form(controlform);
  getsize(&xx,&yy);
  set_bounding_box(0.0,0.0,(float)xx,(float)yy);
  scale_box(&x,&y,&w,&h);
  /* Create the form */
  cur_form = forms[fnumb].form = fl_bgn_form(FL_NO_BOX,w,h);
  fl_end_form();
  forms[fnumb].form->doublebuf = FALSE;
  add_an_object(FL_BOX,FL_UP_BOX,0.0,0.0,w,h);
  /* Get form name and add it */
  strcpy(forms[fnumb].fname,fl_show_input("Give form name:",""));
  fl_add_browser_line(formbrowser,forms[fnumb].fname);
  /* finish off */
  set_form(fnumb++);
  fl_activate_form(controlform);
  changed = 1;
}

void changename_cb(FL_OBJECT *obj, long arg)
/* Callback routine called when the user wants to change the name */
{
  int fn = get_form_numb(cur_form);
  if (cur_form == NULL || fn == -1) return;
  strcpy(forms[fn].fname,fl_show_input("Give form name:",forms[fn].fname));
  fl_replace_browser_line(formbrowser,fn+1,forms[fn].fname);
  changed = 1;
}

void deleteform_cb(FL_OBJECT *obj, long arg)
/* Callback routine called to delete a form */
{
  int i, fn = get_form_numb(cur_form);
  if (cur_form == NULL || fn == -1) return;
  if (! fl_show_question("Delete current form?","","")) return;
  fl_delete_browser_line(formbrowser,fn+1);
  for (i=fn; i<fnumb-1; i++) forms[i] = forms[i+1];
  fnumb--;
  set_form(-1);
  changed = 1;
}

/****
  DRAWING FORMS
****/

void redraw_the_form(int back)
/* Redraws the form in main window. back indicates whether the background
   should be redrawn (when not doublebuffering). This avoids flashing. */
{
  long w,h;
  if (main_window == 0) return;
  winset(main_window);
  getsize(&w,&h);
  if (back)
  {
    reshapeviewport();
    ortho2(-0.5, (float) w-0.5, -0.5, (float) h-0.5);
    fl_color(BLACK); clear();
  }
  if (fl_doublebuf) {fl_color(BLACK); clear();}
  if (cur_form != NULL)
  {
    cur_form->window = (int) main_window;
    cur_form->visible = TRUE;
    fl_redraw_form(cur_form);
    cur_form->window = -1;
    cur_form->visible = FALSE;
    draw_selbox();
  }
  if (fl_doublebuf) swapbuffers();
}

/****
  LOADING AND SAVING
****/

void load_forms(int merge, char *str)
/* loads or merges a file with form definitions */
{
  int i,magic;
  FILE *fn;
  char fname[256], *ff;
  /* Get the filename */
  if (str != NULL && str[0] != '\0')
    ff = str;
  else if (merge)
    ff = fl_show_file_selector("Filename to merge forms from","","*.fd","");
  else
    ff = fl_show_file_selector("Filename to load forms from","","*.fd","");
  if (ff == NULL || ff[0] == '\0') return;
  /* Append .fd if required. */
  strcpy(fname,ff);
  i = strlen(fname) -1;
  if (fname[i]!='d' || fname[i-1]!='f' || fname[i-2]!='.')
    strcat(fname,".fd");
  /* Open the file for reading */
  fn = fopen(fname,"r");
  if (fn == (FILE *) NULL) { fl_show_message("No such file!!","",""); return; }
  /* Read in the definitions */
  if (!merge) { fnumb = 0; fl_clear_browser(formbrowser);}
  magic = 0;
  fscanf(fn,"Magic: %d\n\n",&magic);
  if (magic != 12321) { fl_show_message("Wrong type of file!!","",""); return; }
  fscanf(fn,"Internal Form Definition File\n");
  fscanf(fn,"    (do not change)\n\n");
  fscanf(fn,"Number of forms: %d\n",&i);
  while (i>0 && fnumb <MAXFRM)
  {
    i--;
    read_form(fn,forms[fnumb].fname);
    forms[fnumb].form = cur_form;
    forms[fnumb].form->doublebuf = FALSE;
    fl_add_browser_line(formbrowser,forms[fnumb].fname);
    fnumb++;
  }
  if (i>0) fl_show_message("Not all forms could be loaded!!","","");
  if (i == 0 && !merge)
  {
    fscanf(fn,"\n==============================\n");
    fgets(main_name,64,fn);
    main_name[strlen(main_name)-1] = '\0';
  }
  if (fnumb>0) set_form(0); else set_form(-1);
  if (merge) changed = 1;
  fclose(fn);
}

int save_forms(char *str)
/* saves the form definitions, retunrs whether saved */
{
  int i,j;
  FILE *fn;
  char fname[256], filename[256], *ff;
  /* Get the filename */
  if (str != NULL && str[0] != '\0')
    ff = str;
  else
    ff = fl_show_file_selector("Filename to save forms to","","*.fd","");
  if (ff == NULL || ff[0] == '\0')
    { fl_show_message("No forms were saved.","",""); return 0; }

  /* Remove .fd is required */
  strcpy(filename,ff);
  i = strlen(filename) -1;
  if (filename[i] == 'd' && filename[i-1] == 'f' && filename[i-2] == '.')
    filename[i-2] = '\0';

  /* Make the .h file. */
  strcpy(fname,filename);
  strcat(fname,".h");
  fn = fopen(fname,"w");
  if (fn == (FILE *) NULL)
    { fl_show_message("Cannot create header file!!","",""); return 0; }
  fprintf(fn,"/* Header file generated with fdesign. */\n\n");
  fprintf(fn,"/**** Callback routines ****/\n\n");
  for (i=0; i<fnumb; i++) print_callbacks(fn,forms[i].form);
  fprintf(fn,"\n\n/**** Forms and Objects ****/\n\n");
  for (i=0; i<fnumb; i++) print_objects(fn,forms[i].form,forms[i].fname);
  fprintf(fn,"\n\n/**** Creation Routine ****/\n\n");
  fprintf(fn,"extern void %s(void);\n",main_name);
  fclose(fn);

  /* Make the .c file. */
  strcpy(fname,filename);
  strcat(fname,".c");
  fn = fopen(fname,"w");
  if (fn == (FILE *) NULL)
    { fl_show_message("Cannot create C-code file!!","",""); return 0; }
  fprintf(fn,"/* Form definition file generated with fdesign. */\n\n");
  fprintf(fn,"#include \"forms.h\"\n");
  strcpy(fname,filename);
  j = -1;
  for (i=0; fname[i] != NULL; i++) if (fname[i] == '/') j = i;
  if (j >= 0) for (i=j+1; fname[i-1] != NULL; i++) fname[i-j-1] = fname[i];
  fprintf(fn,"#include \"%s.h\"\n\n",fname);
  for (i=0; i<fnumb; i++) print_form(fn,forms[i].form,forms[i].fname);
  fprintf(fn,"void %s(void)\n{\n",main_name);
  for (i=0; i<fnumb; i++)
     fprintf(fn,"  create_form_%s();\n",forms[i].fname);
  fprintf(fn,"}\n\n");
  fclose(fn);

  /* Make the .fd file. */
  strcpy(fname,filename);
  strcat(fname,".fd");
  fn = fopen(fname,"w");
  if (fn == (FILE *) NULL)
    { fl_show_message("Cannot create definition file!!","",""); return 0; }
  fprintf(fn,"Magic: 12321\n\n");
  fprintf(fn,"Internal Form Definition File\n");
  fprintf(fn,"    (do not change)\n\n");
  fprintf(fn,"Number of forms: %d\n",fnumb);
  for (i=0; i<fnumb; i++) write_form(fn,forms[i].form,forms[i].fname);
  fprintf(fn,"\n==============================\n%s\n",main_name);
  fclose(fn);
  return 1;
}
