/* 
 * fd_rubber.c
 *
 * This file is part of the Form Designer.
 * It contain some routines to make rubberband boxes. It uses the overlay
 * planes for this. This is used for having the user indicate sizes of objects.
 *
 * Written by Mark Overmars
 *
 * Version 2.0a
 * Date  Nov  1, 1991
 */

#include <stdio.h>
#include <gl/gl.h>
#include <gl/device.h>
#include "forms.h"
#include "fd_main.h"

/****
  ROUNDING
*****/

static float xmin = 0.0, ymin = 0.0, width = 1280.0, height = 1024.0;
  /* Bounding box */
static float stepsize = 1.0;	/* Step size (for alignment) */

void set_step_size(float size)
/* Sets the step size */
  { if (size < 1.0) stepsize = 1.0; else stepsize = size; }

float get_step_size()
/* Returns the step size */
  { return stepsize;}

void set_bounding_box(float x,float y,float w,float h)
/* Sets the bounding box. */
  { xmin = x; ymin = y; width = w; height = h; }

static void round_size(float *x,float *y,float *w,float *h)
/* rounds the box size to the stepsize and inside the bounding box */
{
  int t;
  if (*w > width) *w = width;
  if (*h >height) *h = height;
  if (stepsize >0.0)
  {
    t =  (int) (*w / stepsize + 0.5);
    if (*w >= 0.0)  *w = stepsize * t; else *w = stepsize * (t-1);
    t =  (int) (*h / stepsize + 0.5);  *h = stepsize * t;
    if (*h >= 0.0)  *h = stepsize * t; else *h = stepsize * (t-1);
  }
  if (*x + *w > xmin + width)  *w = xmin + width - *x;
  if (*y + *h > ymin + height) *h = ymin + height - *y;
  if (*x + *w < xmin) *w = xmin - *x;
  if (*y + *h < ymin) *h = ymin - *y;
  if (*w >= 0.0 && *w < stepsize) *w = stepsize;
  if (*h >= 0.0 && *h < stepsize) *h = stepsize;
  if (*w < 0.0  && *w > -stepsize) *w = -stepsize;
  if (*h < 0.0  && *h > -stepsize) *h = -stepsize;
}

static void round_position(float *x,float *y,float *w,float *h)
/* rounds the position to the stepsize and inside the bounding box */
{
  int t;
  if (*w > width) *w = width;
  if (*h >height) *h = height;
  if (stepsize >0.0)
  {
    t = (int) (*x / stepsize + 0.5);  *x = stepsize * t;
    t = (int) (*y / stepsize + 0.5);  *y = stepsize * t;
  }
  if (*x < xmin) *x = xmin;
  if (*y < ymin) *y = ymin;
  if (*x + *w > xmin + width)  *x = xmin + width - *w;
  if (*y + *h > ymin + height) *y = ymin + height - *h;
}

/****
  DRAWING
****/

static void show_box(float x,float y,float w,float h)
/* draws a box */
{
  drawmode(PUPDRAW);
  color(1);
  mapcolor(1,255,0,0);
  rect(x,y,x+w-1.0,y+h-1.0);
  drawmode(NORMALDRAW);
}

static void hide_box(float x,float y,float w,float h)
/* hides a box */
{
  drawmode(PUPDRAW);
  color(0);
  rect(x,y,x+w-1.0,y+h-1.0);
  drawmode(NORMALDRAW);
}

/****
  THE ACTUAL ROUTINES
****/

static int ready()
/* Whether we should stop */
{ long dev;
  short val;
  if (! fl_qtest()) return 0;
  if ((dev = fl_qread(&val)) == MOUSE1 || dev == MOUSE2 || dev == MOUSE3)
     return !val;
  return 0;
}

void get_mouse_pos(float *xx,float *yy)
/* returns the position of the mouse in world coordinates*/
{
  long x,y;
  getorigin(&x,&y);
  mapw2(0,(short)(getvaluator(MOUSEX)-x),(short)(getvaluator(MOUSEY)-y),xx,yy);
}

void move_box(float *x,float *y,float *w,float *h,int offset)
/* drag a box around until the user releases a mouse button */
{
  float oldx = *x, oldy = *y, oldw = *w, oldh = *h;
  float xoff, yoff;
  if (offset)
    { get_mouse_pos(&xoff,&yoff); xoff -= *x; yoff -= *y; }
  else
    { xoff = *w/2.0; yoff = *h/2.0; }
  show_box(*x,*y,*w,*h);
  while (! ready())
  {
    get_mouse_pos(x,y);
    *x -= xoff; *y -= yoff;
    round_position(x,y,w,h);
    if (*x != oldx || *y != oldy)
    {
      hide_box(oldx,oldy,oldw,oldh);
      show_box(*x,*y,*w,*h);
      oldx = *x; oldy = *y; oldw = *w; oldh = *h;
    }
  }
  hide_box(oldx,oldy,*w,*h);
}

void scale_box(float *x,float *y,float *w,float *h)
/* scales a box until the user releases a mouse button */
{
  float oldw = 0.0, oldh = 0.0;
  round_position(x,y,w,h);
  show_box(*x,*y,0.0,0.0);
  while (! ready())
  {
    get_mouse_pos(w,h);
    *w -= *x; *h -= *y;
    round_size(x,y,w,h);
    if (*w != oldw || *h != oldh)
    {
      hide_box(*x,*y,oldw,oldh);
      show_box(*x,*y,*w,*h);
      oldw = *w; oldh = *h;
    }
  }
  hide_box(*x,*y,oldw,oldh);
  if (*w < 0.0 ) { *x += *w; *w = -*w; }
  if (*h < 0.0 ) { *y += *h; *h = -*h; }
}
